/* TTF to BMP */
/* Written by Lin YawJen (LYJ) 1994 */
/* edited by Wu LiangSheng (LSN) 1994 */
/* MSC: cl /AL /Ox /G2 demottf.c */
/* GCC: gcc -o ttf2bmp ttf2bmp.c -lpc -lgrx */

/* usage: e.g. compiled as ttf2bmp.exe,
read ntu_li_m.ttf with 256 points (max)

> ttf2bmp ntu_li_m.ttf 256

Hold escape key for a moment to quit.
*/

/*
Input: TTF Font File
Output: font_buf 1 bit per pixel array
To display it on screen, need to write
translation routine for particular graphics mode.
Email me for more information.
Note: some function names may vary between compiler implementations
Try out new ones if you know. Minor changes will be OK.
But don't ask me about your own compiler because I don't know.

Porting tips: use your own Graphics library. The rest may require 
minor changes only.
*/

/* #define NOV -> No Video, will print some info in text */
/* #define NOV */

/* Choose either MSC or GCC */
/* #define MSC */
#define GCC

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <malloc.h>
#include <memory.h>
#include <time.h>

#ifdef MSC
#include <graph.h>
#endif

#ifdef GCC
#define __fastcall
#define __max(a,b) ( a >= b ? a : b )
#define _kbhit kbhit
#define _getch getch
#include <grx.h>
GrContext *gc_buf;
#endif

#define BOOL short
#define FALSE 0
#define TRUE 1
#define EM 1024 /* It is better to get em from file */

struct SUBHD
       {
       unsigned short fstcode, ecnt, ioff;
       short idlt;
       } *subheader = NULL;

typedef short fixp; /* May use long to prevent overflow */
/* Fix point : S_############.## bit, range +-4096 with 1/4 precision */

struct SCANLINE
     {
     short max, cur;
     fixp *grid;
     } *rowarray, *colarray;

/* drawdot and hline mask */
unsigned char mask_pixel[8] = {0x80,0x40,0x20,0x10,8,4,2,1};
unsigned char mask_from[8] = {0xff,0x7f,0x3f,0x1f,0xf,7,3,1};
unsigned char mask_to[8] = {0x80,0xc0,0xe0,0xf0,0xf8,0xfc,0xfe,0xff};

unsigned long cmap, loca, glyf, head, loca_size;
unsigned short subheadkey[256], *glyfarray, *epts_ctr, num_pts;
short num_ctr;
fixp *xcoor, *ycoor;
unsigned char *flag;

void load_table(FILE *fn);
void load_cmap(FILE *fn);
short font_index(unsigned char *str);
BOOL load_font(FILE *fn,short idx);
void init_row();
void convert();
void show_font(short cnt);
void clear_data();

unsigned short get16(FILE *fn)
     {
     unsigned char c1 = getc(fn);
     unsigned char c2 = getc(fn);
     return ((unsigned short)c1)*256+c2;
     }

unsigned long get32(FILE *fn)
     {
     unsigned short s1 = get16(fn);
     unsigned short s2 = get16(fn);
     return ((unsigned long)s1)*65536L+s2;
     }

unsigned char startc[3] = {0xa4,0x40,0};
short ymin, ymax, xmax, xmin, xsize;
#define ysize ymax
short res, xfrwd, yfrwd, pgfrwd;
unsigned char *font_buf, *show_buf;

void main(int argc, char *argv[])
     {
     FILE *fn;
     short cidx, i;
     time_t t1, t2;
     char texnum[10], c;
     int j, bufsiz, k;

     if (argc < 3) { printf("Usage: DEMOTTF ttffile resolution\n"); return; }
     res = atoi(argv[2]);
     if (res > 256) { printf("Resolution must less then 256 pixel\n"); return;}
     /* Restricting because show_buf need four times space for VGA IO */
     fn = fopen(argv[1],"rb");
     if (fn == NULL) { printf("Can't open ttffile\n"); return; }
     load_table(fn);
     load_cmap(fn);
     init_row();
     cidx = font_index(startc);

#ifndef NOV
#ifdef MSC     
     if (!_setvideomode(_SRES16COLOR))
	{ printf("Graphic mode set error\n"); return; }
#endif
#ifdef GCC
     GrSetMode(GR_default_graphics);
     GrClearScreen(0);
     gc_buf = GrCreateContext(xmax, ymax, 0, 0);
#endif
#endif
     
     time(&t1);
     for (i = 0; i < 5401; i++, cidx++)
	 {
	 if (!load_font(fn,cidx))
	    { printf("Font error or composite font\n"); break;}
	 
	 convert();

#ifndef NOV
#ifdef MSC
       show_font(i); 
#endif       
#ifdef GCC
       bufsiz = xsize*ysize;
       memcpy(gc_buf->gc_baseaddr, font_buf, bufsiz); 
       memcpy(gc_buf->gc_baseaddr+bufsiz, font_buf, bufsiz);
       memcpy(gc_buf->gc_baseaddr+2*bufsiz, font_buf, bufsiz);
       memcpy(gc_buf->gc_baseaddr+3*bufsiz, font_buf, bufsiz);
       GrBitBlt(0,50,50,gc_buf,0,0,xmax-1,ymax-1,GrWRITE);  
       sprintf(texnum,"%d",i);
       GrTextXY(0,0,texnum,0,15);
       c=getch();
       if (c==27) break;
#endif
#else
	 printf("%d\n",i);
	 printf("\n**************************\n");
	 bufsiz = xsize*ysize;
	 printf("%d %d %d\n", xsize, ysize, bufsiz);
	 for(j=0; j<bufsiz; j++) { printf("%X",font_buf[j]);
	 if (j%32==0) printf("\n");
	 if (j%640==0) getch();
	 }

#endif
	 clear_data();
if (_kbhit()) if (_getch() == 27) break; 
	 }
     time(&t2);

#ifndef NOV
     _getch();
#ifdef MSC
     _setvideomode(_DEFAULTMODE);
#endif
#ifdef GCC
     GrSetMode(GR_default_text);
#endif
#endif
     
     fclose(fn);
     printf("time %ld\n",t2-t1);
     printf(" xsize %d\n ysize %d\n xmax %d\n ymax %d\n", 
		xsize, ysize, xmax, ymax);

     }

void load_table(FILE *fn)
     {
     char buf[5];
     short i, cnt;

     fseek(fn,4L,SEEK_SET);
     cnt = get16(fn);
     fseek(fn,12L,SEEK_SET);
     for (i = 0; i < cnt; i++)
	 {
	 fread(buf,1,4,fn);
	 buf[4] = 0;
	 if (!strcmp(buf,"cmap")) { get32(fn); cmap = get32(fn); get32(fn); }
	 else if (!strcmp(buf,"loca"))
	      { get32(fn); loca = get32(fn); loca_size = get32(fn); }
	 else if (!strcmp(buf,"glyf"))
	      { get32(fn); glyf = get32(fn); get32(fn); }
	 else if (!strcmp(buf,"head"))
	      { get32(fn); head = get32(fn); get32(fn); }
	 else { get32(fn); get32(fn); get32(fn); } /*skip*/
	 }
     }

void load_cmap(FILE *fn)
     {
     unsigned short i, j, cnt, fmt, len, num_var = 0;
     unsigned long *off;
     fseek(fn,cmap+2,SEEK_SET);
     cnt = get16(fn);
     off = (unsigned long *)calloc(sizeof(unsigned long),cnt);
     for (i = 0; i < cnt; i++) { get16(fn); get16(fn); off[i] = get32(fn); }
     for (i = 0; i < cnt; i++)
	 {
	 fseek(fn,cmap+off[i],SEEK_SET);
	 fmt = get16(fn);
	 if (fmt != 2) continue;
	 len = get16(fn); get16(fn);
	 for (j = 0; j < 256; j++)
	     {
	     subheadkey[j] = get16(fn)/8;
	     if (subheadkey[j] > num_var) num_var = subheadkey[j];
	     }
	 subheader = (struct SUBHD *)calloc(sizeof(struct SUBHD),++num_var);
	 for (j = 0; j < num_var; j++)
	     {
	     subheader[j].fstcode = get16(fn);
	     subheader[j].ecnt = get16(fn);
	     subheader[j].idlt = (short)get16(fn);
	     subheader[j].ioff = get16(fn) - (num_var-j-1)*8-2;
	     }
	 len -= 6+2*256+num_var*8;
	 glyfarray = (unsigned short *)calloc(sizeof(unsigned short), len/2);
	 for (j = 0; j < len/2; j++) glyfarray[j] = get16(fn);
	 break;
	 }
     free(off);
     if (!subheader) { printf("Not found format 2 mapping\n"); exit(-1); }

     /* Get ymax ymin from head */
     fseek(fn,head+36L,SEEK_SET);
     xmin = (short)get16(fn);  /* usual negtive or 0 */
     ymin = (short)get16(fn);
     xmax = (short)get16(fn) - xmin; /* shift to original (0,0) */
     ymax = (short)get16(fn) - ymin;
     xmax = (short)(((long)xmax)*res/EM);
     ymax = (short)(((long)ymax)*res/EM);
     }

void init_row() /* Initial the data space */
     {
     short i;
     xsize = (xmax+7)/8;       /* in byte */
     rowarray = (struct SCANLINE *)calloc(sizeof(struct SCANLINE),ysize);
     if (rowarray == NULL)
	{ printf("Memory error when initial.\n"); exit(1); }
     colarray = (struct SCANLINE *)calloc(sizeof(struct SCANLINE),xmax);
     if (colarray == NULL)
	{ printf("Memory error when initial.\n"); exit(1); }
     for (i = 0; i < ysize; i++)
	 {
	 rowarray[i].max = 32;  /* from max of NTU_MM */
	 rowarray[i].cur = 0;
	 rowarray[i].grid = (fixp *)calloc(sizeof(fixp),32);
	 if (rowarray[i].grid == NULL)
	    { printf("Memory overflow when init rowarray.\n"); exit(2); }
	 }
     for (i = 0; i < xmax; i++)
	 {
	 colarray[i].max = 32;
	 colarray[i].cur = 0;
	 colarray[i].grid = (fixp *)calloc(sizeof(fixp),32);
	 if (colarray[i].grid == NULL)
	    { printf("Memory overflow when init colarray.\n"); exit(2); }
	 }
     font_buf = (unsigned char *)calloc(xsize,ysize+1);
     if (font_buf == NULL)
	{ printf("Memory overflow when init font_buf\n"); exit(3); }
#ifndef NOV
     show_buf = (unsigned char *)calloc(1,xsize*4*ysize+4);
     xfrwd = 100/xsize;
     yfrwd = 600/ysize;
     pgfrwd = xfrwd*yfrwd;
#endif
     }
/* Must free all and allocate new ones when resize the resolution. */

void clear_data()
     {
     short i;
     for (i = 0; i < ysize; i++) rowarray[i].cur = 0;
     for (i = 0; i < xmax; i++) colarray[i].cur = 0;
     memset(font_buf,0,xsize*ysize);
     free(epts_ctr);
     free(flag);
     free(xcoor);
     free(ycoor);
     }

short font_index(unsigned char *str)
      {
      unsigned short idx = subheadkey[str[0]];
      short ridx = subheader[idx].idlt;
      short code = str[1] - subheader[idx].fstcode;

      if (code > subheader[idx].ecnt) return 0; /*Error*/
      ridx += glyfarray[subheader[idx].ioff/2+code];
      return ridx;
      }

BOOL load_font(FILE *fn, short idx)
     {
     long off;
     short i;
     unsigned short j;
     unsigned char c, ct;

     fseek(fn,loca+(long)idx*4L,SEEK_SET);
     off = get32(fn);
     fseek(fn,glyf+off,SEEK_SET);
     num_ctr = (short)get16(fn);
     if (num_ctr < 0) return FALSE;
     get16(fn); get16(fn); get16(fn); get16(fn);
     epts_ctr = (unsigned short *)calloc(sizeof(unsigned short),num_ctr);
     for (i = 0; i < num_ctr; i++) epts_ctr[i] = get16(fn);
     num_pts = epts_ctr[num_ctr-1]+1; /*zero base*/
     j = get16(fn);
     while (j--) getc(fn); /*skip instruction*/
     flag = (unsigned char *)calloc(1,num_pts);
     xcoor = (fixp *)calloc(sizeof(fixp),num_pts);
     ycoor = (fixp *)calloc(sizeof(fixp),num_pts);
     for (j = 0; j < num_pts;)
	 {
	 flag[j++] = c = getc(fn);
	 if ( c&8) { ct = getc(fn); while (ct--) flag[j++] = c; }
	 }
     for (j = 0; j < num_pts; j++)
	 {
	 if (flag[j] & 2)
	    {
	    c = getc(fn);
	    xcoor[j] = (flag[j] & 0x10)?((short)c):(-1*(short)c);
	    }
	 else if (flag[j] & 0x10) xcoor[j] = 0;
	 else xcoor[j] = (short)get16(fn);
	 }
     for (j = 0; j < num_pts; j++)
	 {
	 if (flag[j] & 4)
	    {
	    c = getc(fn);
	    ycoor[j] =(flag[j] & 0x20)?((short)c):(-1*(short)c);
	    }
	 else if (flag[j] & 0x20) ycoor[j] = 0;
	 else ycoor[j] = (short)get16(fn);
	 }
     for (j = 1; j < num_pts; j++)
	 { xcoor[j] += xcoor[j-1]; ycoor[j] += ycoor[j-1]; }
     for (j = 0; j < num_pts; j++)
	 {  /* Scale maybe loose precision in old version */
	 xcoor[j] = (fixp)(((long)(xcoor[j]-xmin))*res/(EM/4))-2; /* Modify */
	 ycoor[j] = (fixp)(((long)(ycoor[j]-ymin))*res/(EM/4))-2;
	 } /*res x res, 1/4 precision, ** must shift coor before scale */
     return TRUE;
     }

fixp lastx, lasty;

#define Moveto(x, y)  {lastx = (x); lasty = (y);}
void Lineto(fixp x, fixp y);
void Curveto(fixp x, fixp y,unsigned short ctrl_fst,unsigned short ctrl_last);
void SplitCurve(fixp x1, fixp y1, fixp x2, fixp y2); /* Modify */
void Draw();

void convert()
     {
     short i;
     unsigned short j, start_offpt, end_offpt, fst;

     for (i = 0, j = 0; i < num_ctr; i++)
	 {
#ifdef NOV
	 printf("%d %d ", xcoor[j], ycoor[j]);
#endif
	 fst = j;
	 Moveto(xcoor[j],ycoor[j]);
	 start_offpt = 0; /*start at least 1*/
	 for (j++; j <= epts_ctr[i]; j++)
	     {
	     if (!(flag[j]&1))
		{ /*Off curve*/
		if (!start_offpt) { start_offpt = end_offpt = j; }
		else end_offpt++;
		}
	     else { /*On Curve*/
		  if (start_offpt)
		     {
		     Curveto(xcoor[j],ycoor[j],start_offpt,end_offpt);
		     start_offpt = 0;
		     }
		  else Lineto(xcoor[j],ycoor[j]);
		  }
	     }
	 if (start_offpt) Curveto(xcoor[fst],ycoor[fst],start_offpt,end_offpt);
	 else Lineto(xcoor[fst],ycoor[fst]);
	 }
     Draw();
     }

void Lineto(fixp x, fixp y)
     {
     short start, end;
     long delta, sy, sx; /* 10 bit point */
     fixp *tmp;

     if (x != lastx) /* usual apply under 96X96 */
	{
	delta = ((long)(y-lasty))*1024/((long)(x-lastx));
	if (x < lastx)
	   {
	   start = (x+3)/4;
	   end = (lastx+3)/4;
	   sy = ((long)y)*256 + (((long)start)*4-((long)x))*delta/4;
	   }
	else {
	     start = (lastx+3)/4;
	     end = (x+3)/4;
	     sy = ((long)lasty)*256 + ((long)(start*4-lastx))*delta/4;
	     }
	if (start < 0) { sy += delta*(-start); start = 0;}
	if (end >= xmax) end = xmax-1;
	for (; start < end; start++, sy += delta)
	    { /* xadd(start,(fixp)(sy/256));*/
	    if (colarray[start].cur == colarray[start].max)
	       { /* Growth 8 each time */
	       tmp = (fixp *)realloc(colarray[start].grid,
				  (colarray[start].max+8)*sizeof(fixp));
	       if (tmp == NULL) { printf("Memory over.\n"); continue; }
	       colarray[start].grid = tmp;
	       colarray[start].max += 8;
	       }
	    colarray[start].grid[colarray[start].cur++] = (fixp)(sy/256);
	    }
	}
     if (y != lasty)
	{
	delta = ((long)(x-lastx))*1024/((long)(y-lasty));
	if (y < lasty)
	   {
	   start = (y+3)/4;
	   end = (lasty+3)/4;
	   sx = ((long)x)*256 + (((long)start)*4-((long)y))*delta/4;
	   }
	else {
	     start = (lasty+3)/4;
	     end = (y+3)/4;
	     sx = ((long)lastx)*256 + ((long)(start*4-lasty))*delta/4;
	     }
	if (start < 0) { sx += delta*(-start); start = 0;}
	if (end >= ymax) end = ymax-1;
	for (; start < end; start++, sx += delta)
	    { /* yadd(start,(fixp)(sx/256));*/
	    if (rowarray[start].cur == rowarray[start].max)
	       { /* Growth 8 each time */
	       tmp = (fixp *)realloc(rowarray[start].grid,
				  (rowarray[start].max+8)*sizeof(fixp));
	       if (tmp == NULL) { printf("Memory over.\n"); continue; }
	       rowarray[start].grid = tmp;
	       rowarray[start].max += 8;
	       }
	    rowarray[start].grid[rowarray[start].cur++] = (fixp)(sx/256);
	    }
	}
     lastx = x; lasty = y;
     }

void Curveto(fixp x, fixp y,unsigned short ctrl_fst,unsigned short ctrl_last)
     {
     unsigned short ctrl_next = ctrl_fst+1;
     fixp _x = xcoor[ctrl_fst], _y = ycoor[ctrl_fst];
     fixp tx, ty;

     Lineto((lastx+_x)/2, (lasty+_y)/2); /*Uniform B-spline*/
     for (; ctrl_fst <= ctrl_last; ctrl_fst++, ctrl_next++)
	 {
	 if (ctrl_next <= ctrl_last)
	    { tx = xcoor[ctrl_next]; ty = ycoor[ctrl_next]; }
	 else { tx = x; ty = y; }
	 SplitCurve(_x, _y, (_x+tx)/2, (_y+ty)/2);
	 _x = tx;
	 _y = ty;
	 }
     Lineto(x, y);
     }

void SplitCurve(fixp x1, fixp y1, fixp x2, fixp y2) /* Great Modify */
     { /*Delta routine for split nonuniform part of B-spline*/
     short n, i, m;
     long p1x = ((long)lastx-2*(long)x1+(long)x2) << 8;
     long p1y = ((long)lasty-2*(long)y1+(long)y2) << 8;
     long ddx, ddy, p2x, p2y, fx, fy, dx, dy;
     long ratio = __max(labs(p1x),labs(p1y)); /* 10 bit point */

     for (n = 0; ratio >= 512L; n++) ratio >>= 2; /* Compute coef. of spline */
     if (n)
	{                                       /* delta = 1/2^n */
	p2x = ((long)x1 - (long)lastx) << 9;
	p2y = ((long)y1 - (long)lasty) << 9;
	ddx = ((p1x*2) >> (2*n)); /* 2 order */
	ddy = ((p1y*2) >> (2*n));
	dx = (p2x >> n)+(p1x >> (2*n));
	dy = (p2y >> n)+(p1y >> (2*n));
	fx = dx + ((long)lastx << 8);
	fy = dy + ((long)lasty << 8);
	for (m = (1 << n), i = 1; i < m; i++)
	    {
	    Lineto((fixp)(fx/256),(fixp)(fy/256));
	    dx += ddx; fx += dx;
	    dy += ddy; fy += dy;
	    }
	}
     Lineto(x2,y2);
     }

short ycur;

/* These two functions use data coordinates */
void Pixel(short x, short y);
void Hline(short x1, short x2);  /* Hline use ycur as y */

void Draw()
     {
     short i, j, k, cnt, len; /* Modify */
     fixp *tmp, t;

     for (i = 0; i < ysize; i++)
	 {
	 if (!rowarray[i].cur) continue;
	 ycur = (ysize-i-1)*xsize;
	 tmp = rowarray[i].grid;
	 cnt = rowarray[i].cur;
	 for (j = 1; j < cnt; j++) /* Insert sort for average length 6*/
	     {
	     for (t = tmp[j], k = j-1; (k >= 0) && (tmp[k] > t); k--)
		 tmp[k+1] = tmp[k];
	     if (k < j-1) tmp[k+1] = t;
	     }
	 cnt &= 0xfffe; /* must be paired */
	 for (j = 0; j < cnt; j+=2)
	     {   /* Scan convert rule 3, usual appling under 96X96 */
	     len = (tmp[j+1]-tmp[j]+2)/4; /* 0.5 */
	     if (len < 2) Pixel(tmp[j]/4,ysize-i-1); /* Modify */
	     else Hline((tmp[j]+3)/4,(tmp[j]+3)/4+len-1);
	     /* Modify use another way do that like old version */
	     }    /* Scan convert rule 0 & 1 */
	 }
     for (i = 0; i < xmax; i++)
	{ /* Only check rule 3, can be skipped when res > 96 */
	if (!colarray[i].cur) continue;
	tmp = colarray[i].grid;
	cnt = colarray[i].cur;
	for (j = 1; j < cnt; j++) /* Insert sort for average length 6*/
	    {
	    for (t = tmp[j], k = j-1; (k >= 0) && (tmp[k] > t); k--)
		tmp[k+1] = tmp[k];
	    if (k < j-1) tmp[k+1] = t;
	    }
	cnt &= 0xfffe; /* must be paired */
	for (j = 0; j < cnt; j+=2) /* Modify */
	    if (tmp[j+1]-tmp[j] < 6) Pixel(i,ysize-1-(tmp[j]+3)/4);
	    /* Scan convert rule 3, usual appling under 96X96 */
	}
     }

void __fastcall Pixel(short x, short y)
     {
     if ((y < 0) || (y >= ysize) || (x < 0) || (x >= xmax)) return;
     font_buf[y*xsize+x/8] |= mask_pixel[x&7];
     }

void Hline(short x1, short x2)
     {
     short _x1, _x2;
     unsigned char *ptr;

     if (x1 < 0)
	{
	if (x2 < 0) return;
	else if (x2 >= xmax) { x1 = 0; x2 = xmax;}
	else x1 = 0;
	}
     else if (x1 >= xmax) return;
     else if (x2 >= xmax) x2 = xmax;
     _x1 = x1/8; _x2 = x2/8;
     ptr = font_buf+ycur+_x1;
     if (_x1 == _x2) *ptr |= (mask_from[x1&7] & mask_to[x2&7]);
     else {
	  *ptr++ |= mask_from[x1&7];
	  for (_x1++; _x1 < _x2; _x1++) *ptr++ = 0xff;
	  *ptr |= mask_to[x2&7];
	  }
     }

#ifdef MSC
void show_font(short cnt)
     {
     unsigned char *ptr = show_buf+4;
     short i;
     for (i = 0; i < ysize; i++ )
	 {
	 memcpy(ptr+xsize*i*4,font_buf+xsize*i,xsize);
	 memcpy(ptr+xsize*i*4+xsize,font_buf+xsize*i,xsize);
	 memcpy(ptr+xsize*i*4+2*xsize,font_buf+xsize*i,xsize);
	 memcpy(ptr+xsize*i*4+3*xsize,font_buf+xsize*i,xsize);
	 }
     cnt %= pgfrwd;
     _putimage(cnt%xfrwd*xsize*8,cnt/xfrwd*ysize,show_buf,_GPSET);
     }
#endif
